use crate::{
    cmd::package::CmdPackage,
    cmd::Cmd,
    config::{from_file as config_from_file, Config},
    log_fatal, return_if,
};
use futures::future::join_all;
use log::{error, info, warn};
use matrix_sdk::{
    deserialized_responses::EncryptionInfo,
    room::{Invited, Room},
    ruma::{
        events::{
            reaction::{ReactionEventContent, Relation},
            room::{
                member::MemberEventContent,
                message::{MessageEventContent, MessageType, TextMessageEventContent},
            },
            AnyMessageEventContent, SyncMessageEvent, SyncStateEvent,
        },
        UserId,
    },
    Client, Result, SyncSettings,
};
use std::{
    convert::TryFrom,
    lazy::SyncLazy,
    process::abort,
    sync::{Arc, Mutex},
};
use tokio::time::{sleep, Duration};

pub struct Bot {
    config: Config,
    client: Option<Client>,
}

static GLOBAL_PKG: SyncLazy<Arc<Mutex<Vec<CmdPackage>>>> =
    SyncLazy::new(|| Arc::new(Mutex::new(Vec::<CmdPackage>::new())));

macro_rules! GLOBAL_PKG_ITER {
    () => {{
        GLOBAL_PKG.lock().unwrap().iter()
    }};
}

macro_rules! GLOBAL_PKG_LOCKED {
    () => {{
        GLOBAL_PKG.lock().unwrap()
    }};
}

async fn join_room(room: &Invited) {
    info!("Autojoining room {}", room.room_id());
    let mut delay = 2; // seconds

    while let Err(err) = room.accept_invitation().await {
        // retry autojoin due to synapse sending invites, before the invited
        // user can join for more information see
        // https://github.com/matrix-org/synapse/issues/4345
        error!(
            "Failed to join room {} ({:?}), retrying in {}s",
            room.room_id(),
            err,
            delay
        );

        sleep(Duration::from_secs(delay)).await;
        delay *= 2;

        if delay > 60 {
            error!("Can't join room {} ({:?})", room.room_id(), err);
            break;
        }
    }
    info!("Successfully joined room {}", room.room_id());
}

async fn on_state_member(
    room_member: SyncStateEvent<MemberEventContent>,
    client: Client,
    room: Room,
) {
    let room_member_state_key = room_member.state_key;
    let own_user_id = client.user_id().await.unwrap();
    if room_member_state_key != own_user_id {
        error!(
            "Room member state key {} don't match bot's id {}",
            room_member_state_key, own_user_id
        );
        return;
    }

    match room {
        Room::Joined(room) => warn!("Room {} already joined", room.room_id()),
        Room::Invited(room) => join_room(&room).await,
        Room::Left(room) => warn!("Leave room {}", room.room_id()),
    };
}

async fn startup_and_log_rooms(client: &Client) {
    for room in client.joined_rooms() {
        info!("Got joined room: <{}>", room.name().unwrap());
    }

    let invited_rooms = client.invited_rooms();
    let list_to_wait_for = invited_rooms.iter().map(|room| {
        warn!("Pending invit on room {}, joining it", room.name().unwrap());
        return join_room(&room);
    });
    join_all(list_to_wait_for).await;
}

async fn on_msg_room_event(
    ev: SyncMessageEvent<MessageEventContent>,
    room: Room,
    __info: Option<EncryptionInfo>,
) {
    let sender_id = ev.sender.clone();
    return_if!(sender_id.as_str() == room.own_user_id().as_str(), ());
    if let (Room::Joined(joined), MessageType::Text(TextMessageEventContent { ref body, .. })) =
        (room, ev.content.msgtype.clone())
    {
        if body.starts_with("%") {
            if let Some(stripped) = body.strip_prefix("%") {
                let args = stripped.split(' ').collect::<Vec<&str>>();
                let (res, emojies) = match GLOBAL_PKG_ITER!().find(|x| x.has_command(&args)) {
                    Some(pkg) => {
                        let (res, emojies) = pkg.handle(&joined, &sender_id, args);
                        (res, emojies.clone())
                    }
                    None => (format!("Command not found"), vec![]),
                };

                let event = ev.into_full_event(joined.room_id().clone());
                let msg = AnyMessageEventContent::RoomMessage(
                    MessageEventContent::text_reply_html(res.clone(), res, &event),
                );
                let send_id = joined.send(msg, None).await.unwrap().event_id;
                join_all(emojies.iter().map(|emoji| {
                    let msg_emoji = AnyMessageEventContent::Reaction(ReactionEventContent::new(
                        Relation::new(send_id.clone(), emoji.clone()),
                    ));
                    joined.send(msg_emoji, None)
                }))
                .await;
            }
        }
    };
}

async fn on_msg_room_reaction(
    ev: SyncMessageEvent<ReactionEventContent>,
    room: Room,
    _: Option<EncryptionInfo>,
) {
    let sender_id = ev.sender.clone();
    return_if!(sender_id.as_str() == room.own_user_id().as_str(), ());
    if let (
        Room::Joined(joined),
        ReactionEventContent {
            relates_to: Relation {
                emoji, event_id, ..
            },
            ..
        },
    ) = (room, ev.content)
    {
        let msg_content = format!(
            "Not handled reactions in room {} with event {event_id}, emoji was {emoji}",
            joined.room_id()
        );
        let msg = AnyMessageEventContent::RoomMessage(MessageEventContent::text_html(
            msg_content.clone(),
            msg_content,
        ));
        joined.send(msg, None).await.unwrap();
    };
}

macro_rules! register_handler_if_enabled {
    ($client: expr, $config: expr, $handler: expr, $handler_name: expr) => {{
        if ($config) {
            info!("Handle {} events", $handler_name);
            ($client).register_event_handler($handler).await;
        } else {
            warn!("Ignore {} events", $handler_name);
        }
    }};
}

impl Bot {
    pub fn new(config_file: String) -> Bot {
        match config_from_file(&config_file) {
            Err(e) => {
                log_fatal!(
                    "Failed to read config file '{}' with error: {}",
                    config_file,
                    e
                );
            }
            Ok(config) => {
                info!("Using config file: {}", config_file);
                Bot {
                    config: config,
                    client: None,
                }
            }
        }
    }

    pub fn get_configuration(&self) -> &Config {
        &self.config
    }

    pub async fn connect(&mut self) -> Result<&mut Self> {
        let alice = UserId::try_from(self.config.user_name.clone())?;
        let client = Client::new(self.config.homeserver_url.clone())?;

        client
            .login(
                alice.localpart(),
                self.config.user_password.as_str(),
                None,
                Some(self.config.display_name.as_str()),
            )
            .await?;

        info!("Logged as: {}", alice);

        client.sync_once(SyncSettings::default()).await?;
        startup_and_log_rooms(&client).await;
        register_handler_if_enabled!(
            client,
            self.config.handle_state_member,
            on_state_member,
            "state member"
        );

        self.client = Some(client);
        return Ok(self);
    }

    pub async fn listen(&self) -> Result<()> {
        match &self.client {
            None => abort(),
            Some(client) => {
                client.register_event_handler(on_msg_room_event).await;
                client.register_event_handler(on_msg_room_reaction).await;

                info!("Entering sync loop");
                let token = client.sync_token().await.unwrap();
                client.sync(SyncSettings::default().token(token)).await;
                Ok(())
            }
        }
    }

    pub fn add_package(config: &Config, name: String, cmds: Vec<Cmd>) {
        let pkg = match config.section(&name) {
            Some(section) => CmdPackage::from_config(&section, cmds),
            None => {
                abort();
            }
        };
        GLOBAL_PKG_LOCKED!().push(pkg);
    }
}
