#![allow(dead_code)]

use crate::{bool_to_result, log_fatal};
use log::warn;
use matrix_sdk::{room::Joined as JoinedRoom, ruma::UserId};
use std::fmt;
use std::result::Result;

#[derive(Clone)]
pub enum CmdAcl {
    Any,
    RoomOnly(String),
    AdminOnly,
    List(Vec<CmdAcl>),
}

#[derive(Clone)]
pub struct CmdAdminRegister {
    admins: Vec<UserId>,
}

impl fmt::Display for CmdAdminRegister {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        if self.admins.len() != 0 {
            write!(
                f,
                "CmdAdminRegister({})",
                self.admins
                    .iter()
                    .map(|x| x.to_string())
                    .reduce(|x1, x2| x1 + ", " + &x2)
                    .unwrap()
            )
        } else {
            write!(f, "CmdAdminRegister()")
        }
    }
}

impl CmdAdminRegister {
    fn is_admin(&self, user_id: UserId) -> bool {
        return self.admins.iter().position(|x| *x == user_id).is_some();
    }

    pub fn add_admin(&mut self, user_id: UserId) {
        if !self.is_admin(user_id.clone()) {
            self.admins.push(user_id);
        }
    }

    pub fn revoke_admin(&mut self, user_id: UserId) -> Result<(), String> {
        match self.admins.iter().position(|x| *x == user_id) {
            Some(index) => {
                self.admins.remove(index);
                return Ok(());
            }
            None => return Err("Is not an admin".to_string()),
        }
    }

    pub fn from_vec(admins_vec: Vec<UserId>) -> CmdAdminRegister {
        CmdAdminRegister { admins: admins_vec }
    }
}

impl fmt::Display for CmdAcl {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            CmdAcl::Any => write!(f, "Any"),
            CmdAcl::AdminOnly => write!(f, "AdminOnly"),
            CmdAcl::RoomOnly(room) => write!(f, "Room{}", room),
            CmdAcl::List(vec) => {
                if vec.len() != 0 {
                    write!(
                        f,
                        "[{}]",
                        vec.iter()
                            .map(|x| x.to_string())
                            .reduce(|x1, x2| x1 + ", " + &x2)
                            .unwrap()
                    )
                } else {
                    write!(f, "[]")
                }
            }
        }
    }
}

impl CmdAcl {
    pub fn action_permited(
        cmd_admin_register: &CmdAdminRegister,
        acls: &CmdAcl,
        room: &JoinedRoom,
        user_id: &UserId,
    ) -> Result<(), String> {
        return match acls {
            CmdAcl::Any => Ok(()),
            CmdAcl::AdminOnly => bool_to_result!(
                cmd_admin_register.is_admin(user_id.clone()),
                (),
                "Is not an admin".to_string()
            ),
            CmdAcl::RoomOnly(room_id) => bool_to_result!(
                room.room_id() == room_id,
                (),
                "Not permited in this room".to_string()
            ),
            CmdAcl::List(acl_list) => {
                if acl_list.iter().any(|acl| match acl {
                    CmdAcl::Any => true,
                    _ => false,
                }) {
                    Ok(())
                } else {
                    match acl_list
                        .iter()
                        .map(|acl| CmdAcl::action_permited(cmd_admin_register, acl, room, user_id))
                        .filter(|acl_res| acl_res.is_err())
                        .map(|acl_res| acl_res.clone().unwrap_err())
                        .reduce(|err1, err2| err1 + ", " + &err2)
                    {
                        Some(err) => Err(err),
                        None => Ok(()),
                    }
                }
            }
        };
    }

    fn from_string_item(string: &str) -> CmdAcl {
        let upper = string.to_uppercase();
        if upper == "ADMINONLY" {
            CmdAcl::AdminOnly
        } else if upper.starts_with("ROOM!") {
            let room = (&string[4..]).to_string();
            CmdAcl::RoomOnly(room)
        } else if upper == "ANY" {
            CmdAcl::Any
        } else {
            log_fatal!("Unknown ACL: {}", string);
        }
    }

    pub fn from_string(string: String) -> CmdAcl {
        let acl_list = string
            .split(",")
            .map(|x| CmdAcl::from_string_item(x.trim()))
            .collect::<Vec<CmdAcl>>();

        if acl_list.len() == 0 {
            warn!("Default initialize CmdAcl, will only be AdminOnly");
            CmdAcl::AdminOnly
        } else {
            CmdAcl::List(acl_list)
        }
    }
}
