#![allow(dead_code)]

pub mod acl;
pub mod package;

use crate::log_fatal;
use std::{boxed::Box, string::String};

trait CmdFnSimple = Fn(Vec<&str>) -> bool + Send + Sync;
trait CmdFnEcho = Fn(Vec<&str>) -> String + Send + Sync;
trait CmdFnMultiEcho = Fn(Vec<&str>) -> Vec<String> + Send + Sync;

enum CmdHandler {
    Simple(Box<dyn CmdFnSimple>),
    Echo(Box<dyn CmdFnEcho>),
    MultiEcho(Box<dyn CmdFnMultiEcho>),
}

pub struct Cmd {
    name: String,
    argc: u32,
    handler: CmdHandler,
    feedback_emojies: Vec<String>,
}

impl Cmd {
    pub fn matches(&self, name: &str, args: &Vec<&str>) -> bool {
        (self.name == *name) && (self.argc as usize == args.len())
    }

    pub fn get_emojies(&self) -> &Vec<String> {
        &self.feedback_emojies
    }

    pub fn exec(&self, args: Vec<&str>) -> String {
        if args.len() != self.argc as usize {
            log_fatal!("Arg count doesn't match");
        }

        match &self.handler {
            CmdHandler::Simple(f) => match f(args) {
                true => "<b>Ok</b>".to_string(),
                false => "<b>Failed!</b>".to_string(),
            },
            CmdHandler::Echo(f) => f(args),
            CmdHandler::MultiEcho(f) => f(args).join("<br>---"),
        }
    }

    pub fn get_name(&self) -> &str {
        &self.name
    }

    pub fn new_simple<F: 'static>(
        name: String,
        argc: u32,
        f: F,
        emojies: Option<Vec<String>>,
    ) -> Cmd
    where
        F: Fn(Vec<&str>) -> bool + Send + Sync,
    {
        Cmd {
            name: name,
            argc: argc,
            handler: CmdHandler::Simple(Box::new(f)),
            feedback_emojies: emojies.or(Some(vec![])).unwrap(),
        }
    }

    pub fn new_echo<F: 'static>(name: String, argc: u32, f: F, emojies: Option<Vec<String>>) -> Cmd
    where
        F: Fn(Vec<&str>) -> String + Send + Sync,
    {
        Cmd {
            name: name,
            argc: argc,
            handler: CmdHandler::Echo(Box::new(f)),
            feedback_emojies: emojies.or(Some(vec![])).unwrap(),
        }
    }

    pub fn new_multi_echo<F: 'static>(
        name: String,
        argc: u32,
        f: F,
        emojies: Option<Vec<String>>,
    ) -> Cmd
    where
        F: Fn(Vec<&str>) -> Vec<String> + Send + Sync,
    {
        Cmd {
            name: name,
            argc: argc,
            handler: CmdHandler::MultiEcho(Box::new(f)),
            feedback_emojies: emojies.or(Some(vec![])).unwrap(),
        }
    }
}
