#![allow(dead_code)]

use crate::log_fatal;
use ini::{Error::Io as IniErrIo, Error::Parse as IniErrParse, Ini, Properties as IniSection};
use log::{error, info, warn, LevelFilter};
use std::{env, path::Path, process::abort, str::FromStr};
use url::Url;

pub struct Config {
    pub display_name: String,  // The display name to set for the bot
    pub homeserver_url: Url,   // The url of the home server, like https://matrix.iiens.net
    pub user_name: String,     // The true user name
    pub user_password: String, // The true user name

    pub handle_state_member: bool, // Install handler for state memver changes
    pub handle_reaction: bool,     // Install handler for emoji reactions

    ini_handler: Ini,
}

impl Config {
    pub fn section(&self, name: &str) -> Option<&IniSection> {
        self.ini_handler.section(Some(name))
    }
}

pub fn setup_logging() {
    // https://docs.rs/log4rs/latest/log4rs/encode/pattern/index.html
    let log_stdout = log4rs::append::console::ConsoleAppender::builder()
        .encoder(Box::new(log4rs::encode::pattern::PatternEncoder::new(
            "{d(%Y-%m-%d %H:%M:%S)} {h({l} - {m})}{n}",
        )))
        .build();
    let log_config = log4rs::config::Config::builder()
        .appender(log4rs::config::Appender::builder().build("stdout", Box::new(log_stdout)))
        .build(
            log4rs::config::Root::builder()
                .appender("stdout")
                .build(LevelFilter::Info),
        )
        .unwrap();
    let _log_handle = log4rs::init_config(log_config).unwrap();
}

pub fn from_file(file_name: &String) -> Result<Config, String> {
    return match Ini::load_from_file(file_name) {
        Err(IniErrIo(e)) => Err(e.to_string()),
        Err(IniErrParse(e)) => Err(e.to_string()),
        Ok(conf) => {
            let hs_url_str = conf.get_from_or(Some("user"), "homeserver", "https://matrix.org");
            let handle_state_member_str = conf
                .get_from_or(Some("handler"), "state_member", "true")
                .to_string();
            let handle_reaction_str = conf
                .get_from_or(Some("handler"), "reaction", "true")
                .to_string();
            return match Url::parse(hs_url_str) {
                Err(e) => Err(e.to_string()),
                Ok(hs_url) => Ok(Config {
                    user_name: conf
                        .get_from_or(Some("user"), "id", "@some-id:matrix.org")
                        .to_string(),
                    user_password: conf
                        .get_from_or(Some("user"), "password", "no-password")
                        .to_string(),
                    display_name: conf
                        .get_from_or(Some("user"), "display_name", "MGB-Hitagi")
                        .to_string(),
                    homeserver_url: hs_url,
                    handle_state_member: <bool as FromStr>::from_str(&handle_state_member_str)
                        .unwrap_or(false),
                    handle_reaction: <bool as FromStr>::from_str(&handle_reaction_str)
                        .unwrap_or(false),
                    ini_handler: conf,
                }),
            };
        }
    };
}

pub fn write_default(file_name: &String) -> Result<(), String> {
    let mut conf = Ini::new();
    conf.with_section(Some("user"))
        .set("id", "@some_id:matrix.org")
        .set("password", "no-password")
        .set("display_name", "MGB-Hitagi");
    conf.with_section(Some("handler"))
        .set("state_member", "false")
        .set("reaction", "true");
    conf.with_section(Some("basic"))
        .set("name", "basic package")
        .set("acl", "any")
        .set("admins", "@some_id:matrix.org");
    match conf.write_to_file_policy(file_name, ini::EscapePolicy::Everything) {
        Ok(()) => {
            info!("Default config file written to {}", file_name);
            return Ok(());
        }
        Err(e) => return Err(e.to_string()),
    };
}

pub fn check_argument_or_abort(should_write_default: bool) {
    if env::args().len() != 2 {
        let default_file = env::args()
            .nth(0)
            .unwrap()
            .split(std::path::MAIN_SEPARATOR)
            .last()
            .unwrap()
            .to_string()
            + ".ini";
        if Path::new(&default_file).exists() {
            log_fatal!(
                "Args count is {}, you need to only specify the config file for the bot",
                env::args().len() - 1
            );
        } else if should_write_default {
            warn!("Try to write the default config file to {}", default_file);
            match write_default(&default_file) {
                Ok(()) => error!(
                    "Args count is {}, you need to only specify the config file for the bot. \
                 The default config file was written to '{}'",
                    env::args().len() - 1,
                    default_file
                ),
                Err(e) => error!(
                    "Args count is {}, you need to only specify the config file for the bot. \
                 The default config file '{}' could not be written: {}",
                    env::args().len() - 1,
                    default_file,
                    e
                ),
            }
            abort();
        }
    }
}
